--[[
ClosetGnome, a World of Warcraft addon to manage item sets.
Modified in WoW 3.1 to just be a LDB plugin, handle keybindings and display the quips.
Copyright (C) 2009 Rabbit.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
]]

local ClosetGnome = CreateFrame("Frame")
_G.ClosetGnome = ClosetGnome
-- Just so older closetgnome macros will still work for people.
function ClosetGnome:WearSet(s) EquipmentManager_EquipSet(s) end
ClosetGnome:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
ClosetGnome.revision = tonumber(string.sub("$Revision: 614 $", 12, -3))

local db = nil
local L = LibStub("AceLocale-3.0"):GetLocale("ClosetGnome")
local icon = LibStub("LibDBIcon-1.0", true)
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local defaultIcon = "Interface\\Icons\\INV_Chest_Cloth_17"
local menuNeedsUpdate = nil
local acd = LibStub("AceConfigDialog-3.0")

local CGLDB = nil
if ldb then
	CGLDB = ldb:NewDataObject("ClosetGnome", {
		type = "data source",
		text = "ClosetGnome",
		icon = defaultIcon,
	})
end

local options = {
	type = "group",
	args = {
		description = {
			type = "description",
			name = L["closetgnome_description"],
			order = 1,
			fontSize = "medium",
		},
		quips = {
			type = "toggle",
			name = L["Quips"],
			desc = L["Toggle outputting random quips when equipping sets."],
			get = function() return ClosetGnome.db.profile.quips end,
			set = function(info, v) ClosetGnome.db.profile.quips = v end,
			order = 20,
			width = "full",
		},
		minimap = {
			type = "toggle",
			name = L["Show minimap icon"],
			desc = L["Toggle showing or hiding the minimap icon."],
			get = function() return not ClosetGnome.db.profile.minimap.hide end,
			set = function(info, v)
				local hide = not v
				ClosetGnome.db.profile.minimap.hide = hide
				if hide then
					icon:Hide("ClosetGnome")
				else
					icon:Show("ClosetGnome")
				end
			end,
			order = 30,
			width = "full",
			hidden = function() return not icon or not CGLDB or not icon:IsRegistered("ClosetGnome") end,
		},
	},
}
local keybindingOptions = {
	type = "group",
	args = {
		kbDesc = { -- Means you can't have a set called kbDesc :p
			type = "description",
			name = L["keybinding_description"],
			order = 101,
		},
	},
}

local NUM_QUIPS = L["num_quips"]
local function getQuip(input)
	local rand = 1
	repeat rand = math.random(1, NUM_QUIPS) until L[rand] ~= nil
	return "|cffeda55f"..L[rand]:format("|cffd9d919"..input.."|cffeda55f").."|r"
end

--------------------------------------------------------------------------------
-- Utility
--

-- Returns the name of the currently equipped set, or nil if none matchs.
-- If one set is included in another set, the set with the higher number of registered slots is returned.
local function guessEquippedSet()
	local currentSlotCount, currentName = 0, nil
	for i = 1, GetNumEquipmentSets() do
		local name = GetEquipmentSetInfo(i)
		local locations = GetEquipmentSetLocations(name)
		if locations then
			local slotCount, slotMatches = 0, 0
			for slot = INVSLOT_FIRST_EQUIPPED, INVSLOT_LAST_EQUIPPED do
				local location = locations[slot]
				if location then
					slotCount = slotCount + 1
					if location == EQUIPMENT_SET_EMPTY_SLOT then
						if not GetInventoryItemID("player", slot) then
							slotMatches = slotMatches + 1
						end
					elseif location ~= EQUIPMENT_SET_ITEM_MISSING then
						local player, bank, bags, actualSlot = EquipmentManager_UnpackLocation(location)
						if player and not bank and not bags and slot == actualSlot then
							slotMatches = slotMatches + 1
						end
					end
				end
			end
			if slotMatches == slotCount and slotCount > currentSlotCount then
				currentName, currentSlotCount = name, slotCount
			end
		end
	end
	return currentName
end

local function openOptions()
	-- Make sure we're expanded so that users can easily see the keybindings
	-- submenu. My own experience is that people have no idea that there
	-- even IS a plus sign next to the addons listed in the sidepane there.
	for i, button in ipairs(InterfaceOptionsFrameAddOns.buttons) do
		if button.element and button.element.name == "ClosetGnome" and button.element.collapsed then
			OptionsListButtonToggle_OnClick(button.toggle)
		end
	end
	InterfaceOptionsFrame_OpenToCategory("ClosetGnome")
end

--------------------------------------------------------------------------------
-- Keybindings
--

local cgKeyFrame = CreateFrame("Button", "ClosetGnomeKeyFrame", UIParent)
cgKeyFrame:SetScript("OnMouseUp", function(self, set) EquipmentManager_EquipSet(set) end)

local function clearBinding(set)
	if not set or not db.keybindings[set] then return end
	SetOverrideBinding(cgKeyFrame, false, db.keybindings[set])
	db.keybindings[set] = nil
	menuNeedsUpdate = true
end

local function registerBinding(combo, set)
	if not combo or not set then return end
	for k, v in pairs(db.keybindings) do
		if k == set then clearBinding(set) end
	end
	db.keybindings[set] = combo
	SetOverrideBindingClick(cgKeyFrame, false, combo, cgKeyFrame:GetName(), set)
	acd:ConfigTableChanged("oOEvent?", "ClosetGnome-Keybindings")
	menuNeedsUpdate = true
end

--------------------------------------------------------------------------------
-- Events
--

function ClosetGnome:VARIABLES_LOADED()
	SetCVar("equipmentManager", 1)
	GearManagerToggleButton:Show()
end

function ClosetGnome:ADDON_LOADED(addon)
	if addon:lower() ~= "closetgnome" then return end

	self.db = LibStub("AceDB-3.0"):New("ClosetGnomeDB", {
		profile = {
			quips = true,
			minimap = {
				hide = false,
			},
		},
		char = {
			keybindings = {},
		},
	})
	db = self.db.char

	LibStub("AceConfig-3.0"):RegisterOptionsTable("ClosetGnome", options)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ClosetGnome-Keybindings", keybindingOptions)
	acd:AddToBlizOptions("ClosetGnome", "ClosetGnome")
	acd:AddToBlizOptions("ClosetGnome-Keybindings", L["Keybindings"], "ClosetGnome")
	SlashCmdList["CLOSETGNOME"] = openOptions
	SLASH_CLOSETGNOME1 = "/closetgnome"
	SLASH_CLOSETGNOME2 = "/cg"

	-- Create the button frames for keybindings
	for k, v in pairs(db.keybindings) do
		registerBinding(v, k)
	end
	if icon and CGLDB then
		icon:Register("ClosetGnome", CGLDB, self.db.profile.minimap)
	end

	-- AddonLoader support
	if IsLoggedIn() then self:PLAYER_LOGIN() end
end


function ClosetGnome:EQUIPMENT_SETS_CHANGED()
	menuNeedsUpdate = true

	local counter = 150
	for k, v in pairs(keybindingOptions.args) do
		if k ~= "kbDesc" then
			v.hidden = true
		end
	end

	local sets = {}
	for i = 1, GetNumEquipmentSets() do
		local name, icon = GetEquipmentSetInfo(i)
		sets[name] = true
		if keybindingOptions.args[name] then
			keybindingOptions.args[name].hidden = false
			keybindingOptions.args[name .. "_label"].hidden = false
			keybindingOptions.args[name .. "_spacer"].hidden = false
		else
			keybindingOptions.args[name .. "_label"] = {
				type = "description",
				name = "",
				width = "half",
			}
			keybindingOptions.args[name] = {
				type = "keybinding",
				name = name,
				desc = L["Assign a keybinding to %s."]:format(name),
				get = function() return db.keybindings[name] end,
				set = function(info, value)
					if not value or value:len() < 1 then
						clearBinding(name)
					else
						registerBinding(value, name)
					end
				end,
			}
			keybindingOptions.args[name .. "_spacer"] = {
				type = "description",
				name = "",
			}
		end
		keybindingOptions.args[name .. "_label"].order = counter
		keybindingOptions.args[name .. "_label"].image = icon or defaultIcon
		keybindingOptions.args[name].order = counter + 1
		keybindingOptions.args[name .. "_spacer"].order = counter + 2
		counter = counter + 10
	end
	for k in pairs(db.keybindings) do
		if not sets[k] then clearBinding(set) end
	end
	wipe(sets)
end

function ClosetGnome:PLAYER_LOGIN()
	local name = guessEquippedSet()
	if name then self:EQUIPMENT_SWAP_FINISHED(true, name, true) end
	self:EQUIPMENT_SETS_CHANGED()
end

function ClosetGnome:EQUIPMENT_SWAP_FINISHED(success, name, silent)
	if not success then return end
	if CGLDB then
		if name then
			CGLDB.text = name
			local icon = GetEquipmentSetInfoByName(name)
			-- Work around a strange Blizzard API inconsistency where
			-- GetEquipmentSetInfoByName only returns the actual icon name, NOT
			-- prepended by Interface\\Icons\\.
			if icon and icon:sub(1, 9) ~= "Interface" then icon = "Interface\\Icons\\" .. icon end
			CGLDB.icon = icon or defaultIcon
		else
			CGLDB.text = "ClosetGnome"
			CGLDB.icon = defaultIcon
		end
	end

	if not silent and name and ClosetGnome.db.profile.quips then
		DEFAULT_CHAT_FRAME:AddMessage(getQuip(name))
	end
end

--------------------------------------------------------------------------------
-- LDB Plugin
--

if CGLDB then
local popupFrame = CreateFrame("Frame", "ClosetGnomeMenu", UIParent, "UIDropDownMenuTemplate")
local menu = {}

local function menuSorter(a, b)
	return a.text > b.text
end

local function updateMenu()
	menuNeedsUpdate = nil
	menu = wipe(menu)
	for i = 1, GetNumEquipmentSets() do
		local name, icon = GetEquipmentSetInfo(i)
		local k = name
		if db.keybindings[k] then
			name = name .. " |cff999999(" .. db.keybindings[k] .. ")|r"
		end
		local x = {
			text = name,
			tooltipTitle = k,
			tooltipText = string.format("|cffeda55fClick|r to equip or |cffeda55fCtrl-Click|r to delete %s.", k),
			func = function()
				if IsControlKeyDown() then
					DeleteEquipmentSet(k)
				else
					EquipmentManager_EquipSet(k)
				end
			end,
			icon = icon or nil,
		}
		table.insert(menu, x)
	end
	table.sort(menu, menuSorter)
end

function CGLDB.OnClick(self, button)
	if button == "RightButton" then
		openOptions()
	else
		if menuNeedsUpdate then updateMenu() end
		EasyMenu(menu, popupFrame, self, 20, 4, "MENU")
	end
end

function CGLDB.OnTooltipShow(tt)
	tt:AddLine("ClosetGnome")
	tt:AddLine("|cffeda55fClick|r to access the menu, where you can equip sets or delete them by clicking them while holding down Ctrl. |cffeda55fRight-Click|r to pop open the ClosetGnome interface options.", 0.2, 1, 0.2, 1)
end

end

--------------------------------------------------------------------------------
-- Register events
--

ClosetGnome:RegisterEvent("ADDON_LOADED")
ClosetGnome:RegisterEvent("EQUIPMENT_SETS_CHANGED")
ClosetGnome:RegisterEvent("PLAYER_LOGIN")
ClosetGnome:RegisterEvent("VARIABLES_LOADED")
ClosetGnome:RegisterEvent("EQUIPMENT_SWAP_FINISHED")

